#include <llvm/Support/IRBuilder.h>
#include <llvm/ModuleProvider.h>
#include <llvm/Support/MemoryBuffer.h>
#include <llvm/Bitcode/ReaderWriter.h>
#include <llvm/Module.h>
#include <llvm/Function.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/GenericValue.h>
#include <assert.h>
#include <stdio.h>

using namespace llvm;

int main()
{
	// init stuff and load module.bc file
	Module *module;
	MemoryBuffer* buf;
	std::string err;

	if (!(buf = MemoryBuffer::getFile("module.bc", &err))) {
		fprintf(stderr, "Couldn't read handlers file: %s\n", err.c_str());
		return -1;
	}

	if (!(module = llvm::ParseBitcodeFile(buf, &err))) {
		fprintf(stderr, "Couldn't parse handlers file: %s\n", err.c_str());
		return -1;
	}

	ModuleProvider* provider = new ExistingModuleProvider(module);
	assert(provider);

	ExecutionEngine* engine = ExecutionEngine::create(provider);
	assert(engine);



	// first JIT compile some function
	Function* Fn = module->getFunction("get_handler");
	engine->getPointerToFunction(Fn);


	// then create some global variable.
	// this will be dumped to the address after the end of Fn, which is
	// an unallocated memory block
	//
	// the value is important. the less significant bit must be one. if zero,
	// the problem goes undetected
	GlobalVariable* GV = new GlobalVariable(Type::Int32Ty, true, GlobalValue::InternalLinkage, ConstantInt::get(Type::Int32Ty, 0xffffffff), "myInt", module);
	engine->getOrEmitGlobalVariable(GV);


	// and now compile some function
	// the JIT memory engine will try to use the same memory as used by the GV
	// it will fail the assertion because we have replaced the ThisAllocated bit with 1
	Function* Fn2 = module->getFunction("my_function");

	// fails with:
	// test: JITMemoryManager.cpp:127: <unnamed>::FreeRangeHeader*<unnamed>::FreeRangeHeader::AllocateBlock():
	// Assertion `!ThisAllocated && !getBlockAfter().PrevAllocated && "Cannot allocate an allocated block!"' failed.
	engine->getPointerToFunction(Fn2);


	assert(0 && "should had crashed by now!");

	return 0;
}
